/*
 * (C) 1997,2003  Jens Ch. Restemeier <jrestemeier@currantbun.com>
 *
 * (C) 2008       Torsten Neuer <tneuer@inwise.de>
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *
 *  $Id: uf-pcode.h,v 1.8 2008/06/16 09:38:43 tneuer Exp $
 *
 *  $Log: uf-pcode.h,v $
 *  Revision 1.8  2008/06/16 09:38:43  tneuer
 *  Bugfix: 'z' parameter in source function must be unsigned
 *  Added channel variants of source function
 *
 *  Revision 1.7  2008/06/13 12:43:39  tneuer
 *  Use global constants for limits of temporary storage
 *
 *  Revision 1.6  2008/06/13 11:06:16  tneuer
 *  Added range-check optimized variants of get() / put() functions (get0 (), put0 ())
 *
 *  Revision 1.5  2008/06/11 17:52:43  tneuer
 *  Removed assertions not required anymore.
 *  Pointers to const data are const now, too.
 *
 *  Revision 1.4  2008/06/07 14:55:48  tneuer
 *  New cosine and tangens tables to increase accuracy
 *
 *  Revision 1.3  2008/05/29 11:46:20  tneuer
 *  Minor code cleanup.
 *
 *  Revision 1.2  2008/05/21 12:30:47  tneuer
 *  Minor code cleanup.
 *
 *  Revision 1.1.1.1  2008/04/27 13:19:02  tneuer
 *  initial import
 *
 *  Revision 1.1  2008/04/16 14:26:59  tneuer
 *  Initial revision
 *
 */

#ifndef __UF_PCODE_H__
#define __UF_PCODE_H__

#include <stdlib.h>

#include <glib/gmacros.h>

#include <libgimp/gimp.h>

#include "cos-tab.h"
#include "tan-tab.h"


/*
 *  retrieve value for a pixel at (xx,yy) in channel (zz)
 */
static inline gint
p_src (const gint xx, const gint yy, const guint zz)
{
  return drawable_src (xx, yy, zz);
}


static inline gint
p_src_r (const gint xx, const gint yy)
{
  return drawable_src_r (xx, yy);
}


static inline gint
p_src_g (const gint xx, const gint yy)
{
  return drawable_src_g (xx, yy);
}


static inline gint
p_src_b (const gint xx, const gint yy)
{
  return drawable_src_b (xx, yy);
}


static inline gint
p_src_a (const gint xx, const gint yy)
{
  return drawable_src_a (xx, yy);
}


/*
 *  
 */
static inline gint
p_cnv (const gint m11, const gint m12, const gint m13,
       const gint m21, const gint m22, const gint m23,
       const gint m31, const gint m32, const gint m33, const gint dd)
{
  return ((dd != 0) ?
	  (m11 * p_src (environment.x - 1, environment.y - 1, environment.z) +
	   m12 * p_src (environment.x,     environment.y - 1, environment.z) +
	   m13 * p_src (environment.x + 1, environment.y - 1, environment.z) +
	   m21 * p_src (environment.x - 1, environment.y,     environment.z) +
	   m22 * p_src (environment.x,     environment.y,     environment.z) +
	   m23 * p_src (environment.x + 1, environment.y,     environment.z) +
	   m31 * p_src (environment.x - 1, environment.y + 1, environment.z) +
	   m32 * p_src (environment.x,     environment.y + 1, environment.z) +
	   m33 * p_src (environment.x + 1, environment.y + 1, environment.z)) / dd : 255);
}


/*
 *  return value of slider (i)
 */
static inline gint
p_ctl (const gint i)
{
  /*
   *  FIXME:  we should emit a run-time error (out of bounds)
   *  for illegal values instead of simply returning 0.
   */
  return ((i < 0) || (i > 7)) ? 0 : environment.value[i];
}


/*
 *  return the value of slider (i) mapped onto the range (a..b)
 */
static inline gint
p_val (const gint i, const gint a, const gint b)
{
  return ((p_ctl (i) * (b - a)) / 255) + a;
}


/*
 *  return the absolute value of the argument (a)
 */
static inline gint
p_abs (const gint a)
{
  return ABS (a);
}


/*
 *  return absolute value of difference of (a) and (b)
 */
static inline gint
p_dif (const gint a, const gint b)
{
  return p_abs (a - b);
}


/*
 *  store value (v) in temporary register (i)
 */

static inline gint
p_put0 (const gint v, const gint i)
{
  /*
   *  FIXME:  we should emit a run-time error (out of bounds)
   *  for illegal values instead of simply returning 0.
   */
  return (environment.storage[i] = v);
}

static inline gint
p_put (const gint v, const gint i)
{
  /*
   *  FIXME:  we should emit a run-time error (out of bounds)
   *  for illegal values instead of simply returning 0.
   */
  return ((i >= 0) && (i < N_TEMPSLOTS)) ? p_put0(v, i) : 0;
}


/*
 *  retrieve value from temporary register (i)
 */
static inline gint
p_get0 (const gint i)
{
  /*
   *  FIXME:  we should emit a run-time error (out of bounds)
   *  for illegal values instead of simply returning 0.
   */
  return environment.storage[i];
}

static inline gint
p_get (const gint i)
{
  /*
   *  FIXME:  we should emit a run-time error (out of bounds)
   *  for illegal values instead of simply returning 0.
   */
  return ((i >= 0) && (i < N_TEMPSLOTS)) ? p_get0(i) : 0;
}


/*
 * This is a little unsusual: Normally you'll create a sinus-tab, and do
 * a cosinus by adding an offset.
 * But this should be the easiest way to allow negative parameters: As
 * cosinus is symmetric at x=0 ( cos(x)==cos(-x) ) I can use an abs(x).
 *
 * Thanks to Michael Johannhanwahr for these tables.
 */

static inline gint
p_cos (const gint i)
{
  return cos_tab[p_abs (i) & (COS_TAB_SIZE - 1)];
}


static inline gint
p_sin (const gint i)
{
  return p_cos (i - (COS_TAB_SIZE / 4));
}


static inline gint
p_tan (const gint i)
{
  return tan_tab[(i + (TAN_TAB_SIZE / 2)) & (TAN_TAB_SIZE - 1)];
}


/*
 *  convert radial coordinate to cartesian x
 */
static inline gint
p_r2x (const gint dd, const gint mm)
{
  return (cos (M_PI * (float) dd / (COS_TAB_SIZE / 2.0)) * (float) mm);
}


/*
 *  convert radial coordinate to cartesian y
 */
static inline gint
p_r2y (const gint dd, const gint mm)
{
  return (sin (M_PI * (float) dd / (COS_TAB_SIZE / 2.0)) * (float) mm);
}


/*
 *  retrieve value for a pixel in channel (zz) which is (mm) units
 *  away at an angle of (dd) from the center of the image
 */
static inline gint
p_rad (const gint dd, const gint mm, const gint zz)
{
  return p_src (environment.Xd2 - p_r2x (dd, mm),
		environment.Yd2 - p_r2y (dd, mm), zz);
}


/*
 *  return square root of value (i)
 */
static inline gint
p_sqr (const gint i)
{
  return (i > 0) ? sqrt (i) : 0;
}


static inline gint
p_div (const gint a, const gint b)
{
  return (b != 0) ? a / b : 1;
}


static inline gint
p_mod (const gint a, const gint b)
{
  return (b != 0) ? a % b : 0;
}


/*
 *  return the lesser of values (a) and (b)
 */
static inline gint
p_min (const gint a, const gint b)
{
  return MIN (a, b);
}


/*
 *  return the bigger of values (a) and (b)
 */
static inline gint
p_max (const gint a, const gint b)
{
  return MAX (a, b);
}


/*
 *  return a randon number in range (a..b) inclusive.
 *
 *  "... returns a different number each time it is called, but
 *  the entire function resets each time an image is processed.
 *  As a result, a filter that uses the rnd function will have
 *  the same effect each time it is used on the same image."
 *
 *  Which means, that the random number seed value need of course
 *  to be consistent at least during a single run of GIMP.
 */
static inline gint
p_rnd (const gint a, const gint b)
{
  return ( (a == b) ?
	   a :
	   ( (a < b) ?
	     g_rand_int_range (environment.rand, a, b) :
	     g_rand_int_range (environment.rand, b, a) ) );
}


/*
 *  map a value (a) from an input range (il..ih)
 *  onto an output range (ol..oh)
 */
static inline gint
p_scl (const gint a,
       const gint il, const gint ih,
       const gint ol, const gint oh)
{
  return (ih != il) ? (((a - il) * (oh - ol)) / (ih - il)) + ol : 0;
}


/*
 *  retrieve an item (i) from a mapping table (n)
 *
 *  the mapping tables are computed from slider pairs
 *  0:(0,1),1:(2,3),2:(4,5),3:(6,7)
 *
 *  NOTE:  The function could be sped up by pre-computing
 *  all the values and storing them in the environment (e)
 *  in function calc_envir().  However, this would prove
 *  bad for any future extensions to user filter, since
 *  for each channel value there needs to be a map entry
 *  defined - which is 2^n entries for n-bit wide channels.
 */
static inline gint
p_map (const gint n, const gint i)
{
  /*
   *  FIXME: There should rather be a run-time error
   *  for "illegal map index" instead of returning 0.
   */
  /*
   *  Taken from the FF Documentation:
   *
   *  "Each table uses a pair of sliders: table n uses sliders
   *  2n and 2n+1 for the high and low values, respectively.
   *  The table is constructed as follows, where L is the value
   *  of the low slider, H is the value of the high slider, and
   *  I is an entry: if I<=L use 0; if I>=H use 255; if L<I<H,
   *  use (I-L)*255/(H-L)".
   */
  const gint a2 = (n << 1);
  const gint h = p_ctl (a2);
  const gint l = p_ctl (a2 + 1);

  if (i <= l)
    return 0;
  if (i >= h)
    return 255;
  /* l < i < h */
  return (i - l) * 255 / (h - l);

#if 0
  /*
   *  This is the original formula used by User Filter v0.8.
   */
  const gint a2 = (n << 1);
  const gint x = p_ctl (a2);
  const gint y = p_ctl (a2 + 1);

  return p_abs ((i * (y - x) / 255) + x);
#endif
}


static inline gint
p_add (const gint a, const gint b, const gint c)
{
  return p_min (a + b, c);
}


static inline gint
p_mix (const gint a, const gint b, const gint n, const gint d)
{
  return (d != 0) ? a * n / d + b * (d - n) / d : 0;
}


static inline gint
p_sub (const gint a, const gint b, const gint c)
{
  return p_max (p_dif (a, b), c);
}


/*
 *  convert cartesian coordinate to radial coordinate (angle)
 */
static inline gint
p_c2d (const gint x, const gint y)
{
  /*
   *  This has been taken from Filter Foundry,
   *  http://www.telegraphics.com.au/sw/info/filterfoundry.html
   *  (Code released under GPL)
   */
  return (atan2 (-y, -x) * TAN_TAB_SIZE / M_PI);
#if 0
  /*
   *  This is the code used in User Filter < v0.9.7
   */
  const gint d = ((y != 0) ?
		  (256.0 *
		   (atan ((float) x / (float) p_abs (y)) / (M_PI / 2.0))) -
		  256 : (x < 0) ? 512 : 0);
  return (y < 0) ? d : -d;
#endif
}


/*
 *  convert cartesian coordinate to radial coordinate (distance)
 */
static inline gint
p_c2m (const gint x, const gint y)
{
  return p_sqr (x * x + y * y);
}


#endif /* __UF_PCODE_H__ */
