/*
 * (C) 1997,2003  Jens Ch. Restemeier <jrestemeier@currantbun.com>
 *
 * (C) 2008       Torsten Neuer <tneuer@inwise.de>
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *
 *  $Id: userfilter.h,v 1.10 2008/06/18 09:08:16 tneuer Exp $
 *
 *  $Log: userfilter.h,v $
 *  Revision 1.10  2008/06/18 09:08:16  tneuer
 *  Add reset button for edge mode in control panel.
 *
 *  Revision 1.9  2008/06/17 14:43:45  tneuer
 *  Added editor component for default edge mode
 *
 *  Revision 1.8  2008/06/13 12:45:50  tneuer
 *  Add support for different edge modes for the pixel fetcher
 *
 *  Revision 1.7  2008/06/11 17:52:52  tneuer
 *  Removed assertions not required anymore.
 *  Pointers to const data are const now, too.
 *
 *  Revision 1.6  2008/06/06 13:03:55  tneuer
 *  No more help button when running filters in stand-alone mode
 *
 *  Revision 1.5  2008/05/28 10:01:20  tneuer
 *  Using GtkPaned to divide preview and controls
 *
 *  Revision 1.4  2008/05/26 13:03:20  tneuer
 *  Only display random seed reset widget if filter uses rnd() function
 *
 *  Revision 1.3  2008/05/21 12:30:50  tneuer
 *  Minor code cleanup.
 *
 *  Revision 1.2  2008/05/19 10:55:23  tneuer
 *  Code cleanup: merge duplicate code in plug-in dialog(s)
 *
 *  Revision 1.1.1.1  2008/04/27 13:19:05  tneuer
 *  initial import
 *
 *  Revision 1.1  2008/04/16 11:28:09  tneuer
 *  Initial revision
 *
 */

#ifndef __USERFILTER_H__
#define __USERFILTER_H__

#include <stdio.h>
#include <glib/gtypes.h>
#include <libgimp/gimp.h>
#include <libgimp/gimpui.h>

#include "guf.h"


enum {
  N_MAPS      =   4, /* # of map() values */
  N_CONTROLS  =   8, /* # of ctl() values */
  N_CHANNELS  =   4, /* # of color channels */
  N_TEMPSLOTS = 256  /* # of get () / put () storage slots */
};


/*
 *  FIXME:  We have to make this dynamic again if we want to make it thread-safe.
 *          Planned for 0.9.4.
 */

typedef struct FILTER_ENVIRONMENT
{
  /*
   *  FF elements...
   */
  gint storage[N_TEMPSLOTS];	/* 256 values storage for put()+get() functions */
  gint * value;                 /* 8 int slider-values (shadow control.value) */
  gint r;			/* colors in RGB space... */
  gint g;
  gint b;
  gint a;			/* alpha at pixel postion */
  gint c;			/* current color (=src(x,y,z)) */
  gint i;			/* colors in YUV space... */
  gint u;
  gint v;
  gint x;			/* x position */
  gint y;			/* y position */
  gint z;			/* current channel: 0=red, 1= green, 2=blue, 3=alpha */
  gint X;			/* X-width of image (width) */
  gint Y;			/* Y-width of image ( height) */
  gint Z;			/* number of channels of image (1: Gray, 2: GrayA, 3: RGB, 4: RGBA) */
  gint d;			/* direction from center of the pic (-512 to 512) */
  gint m;			/* distance to center of the pic */
  gint M;			/* maximum distance from the center */
  /*
   *  FF elements used in optimizations...
   */
  gint Xm1;			/* maximum x coordinate: X - 1 */
  gint Ym1;			/* maximum y coordinate: Y - 1 */
  gint Xd2;			/* center x coordinate:  X / 2 */
  gint Yd2;			/* center y coordinate:  Y / 2 */
  /*
   *  non-FF elements...
   */
  GRand * rand;
  GimpDrawable * drawable;	/* the current drawable */
  guint32 drawable_hash;	/* hash id for this image */
} FilterEnvironment;


typedef struct FILTER_VALUES
{
  gint                     value[N_CONTROLS]; /* current control values used */
  GimpPixelFetcherEdgeMode edge_mode;         /* edge mode used for the pixel fetcher */
} FilterValues;


extern FilterEnvironment environment;  /* pixel environment of the renderer */
extern UserFilter *      filter;       /* the filter data */
extern FilterValues      control;

extern const gchar * const plugin_name;/* "plug-in-" */


/*
 *  This is a generic dialog for all user filters which are
 *  running in stand-alone mode (either through the user filter
 *  plugin or as compiled native plugin).
 */

gboolean userfilter_dialog (GimpDrawable * drawable);


#if __USERFILTER_INTERNAL__
/*
 *  "Private" stuff that is shared with the plug-in's common interface
 *  but shouldn't be visible to anything else...
 */

extern GtkWidget * dialog_dialog;
extern GtkWidget * dialog_preview;
extern GtkWidget * dialog_edge;
extern GtkWidget * dialog_rnd;
extern GtkWidget * dialog_sreset[N_CONTROLS];
extern GtkObject * dialog_slider[N_CONTROLS];


void _uf_slider_hide (const guint slider);

void _uf_build_page_filter_control (const GtkWidget * notebook);

GtkWidget * _uf_build_preview (GtkPaned *     box,
			       GimpDrawable * drawable);

GtkWidget * _uf_notebook_tab (const GtkWidget *   notebook,
			      const gchar * const icon,
			      const gchar * const text,
			      const gchar * const tooltip);

gboolean _filter_uses_rnd (const UserFilter * const filter);

#endif

#endif /* __USERFILTER_H__ */
